from time import sleep

from requests import HTTPError

from stocra.models import StocraHTTPError
from stocra.utils import calculate_sleep


def retry_on_service_unavailable(error: StocraHTTPError) -> bool:
    if not isinstance(error.exception, HTTPError):
        return False

    if error.exception.response.status_code != 503:
        return False

    if error.iteration > 10:
        return False

    sleep(calculate_sleep(error.iteration))
    return True


def retry_on_too_many_requests(error: StocraHTTPError) -> bool:
    if not isinstance(error.exception, HTTPError):
        return False

    if error.exception.response.status_code != 429:
        return False

    if error.iteration > 10:
        return False

    sleep(int(error.exception.response.headers["Retry-After"]))
    return True



import asyncio

from aiohttp import ClientConnectionError, ClientResponseError

from stocra.models import StocraHTTPError
from stocra.utils import calculate_sleep


async def retry_on_service_unavailable(error: StocraHTTPError) -> bool:
    if not isinstance(error.exception, ClientResponseError):
        return False

    if error.exception.status != 503:
        return False

    if error.iteration > 10:
        return False

    await asyncio.sleep(calculate_sleep(error.iteration))
    return True


async def retry_on_connection_error(error: StocraHTTPError) -> bool:
    if not isinstance(error.exception, ClientConnectionError):
        return False

    if error.iteration > 10:
        return False

    await asyncio.sleep(calculate_sleep(error.iteration))
    return True


async def retry_on_too_many_requests(error: StocraHTTPError) -> bool:
    if not isinstance(error.exception, ClientResponseError):
        return False

    if error.exception.status != 429:
        return False

    if error.iteration > 10:
        return False

    headers = error.exception.headers

    if not headers:
        return False

    retry_after = headers.get("Retry-After", 0)
    await asyncio.sleep(int(retry_after))
    return True



import logging
from concurrent.futures import Executor, as_completed
from decimal import Decimal
from itertools import count
from time import sleep
from typing import Dict, Iterable, List, Optional, Tuple, Union, cast

from requests import HTTPError, RequestException, Session

from stocra.base_client import StocraBase
from stocra.models import Block, ErrorHandler, StocraHTTPError, Token, Transaction

logger = logging.getLogger("stocra")


class Stocra(StocraBase):
    _session: Session
    _executor: Optional[Executor]

    def __init__(
        self,
        api_key: Optional[str] = None,
        session: Optional[Session] = None,
        executor: Optional[Executor] = None,
        error_handlers: Optional[List[ErrorHandler]] = None,
    ):
        super().__init__(
            api_key=api_key,
            error_handlers=error_handlers,
        )
        self._session = session or Session()
        self._executor = executor

    def get_block(self, blockchain: str, hash_or_height: Union[str, int] = "latest") -> Block:
        logger.debug("%s: get_block %s", blockchain, hash_or_height)
        block_json = self._get(blockchain=blockchain, endpoint=f"blocks/{hash_or_height}")
        return Block(**block_json)

    def get_transaction(self, blockchain: str, transaction_hash: str) -> Transaction:
        logger.debug("%s: get_transaction %s", blockchain, transaction_hash)
        transaction_json = self._get(blockchain=blockchain, endpoint=f"transactions/{transaction_hash}")
        return Transaction(**transaction_json)

    def get_all_transactions_of_block(self, blockchain: str, block: Block) -> Iterable[Transaction]:
        logger.debug("%s: get_all_transactions %s", blockchain, block.height)
        if self._executor:
            futures = [
                self._executor.submit(self.get_transaction, blockchain, transaction_hash)
                for transaction_hash in block.transactions
            ]
            for transaction_task in as_completed(futures):
                yield transaction_task.result()
        else:
            for transaction_hash in block.transactions:
                yield self.get_transaction(blockchain, transaction_hash)

    def stream_new_blocks(
        self,
        blockchain: str,
        start_block_hash_or_height: Union[int, str] = "latest",
        sleep_interval_seconds: float = 10,
    ) -> Iterable[Block]:
        block = self.get_block(blockchain=blockchain, hash_or_height=start_block_hash_or_height)
        next_block_height = block.height + 1
        yield block

        while True:
            try:
                block = self.get_block(blockchain=blockchain, hash_or_height=next_block_height)
            except HTTPError as exception:
                if exception.response.status_code == 404:
                    self._handle_404_during_block_streaming(blockchain, next_block_height, sleep_interval_seconds)
                    continue

                raise

            next_block_height += 1
            yield block

    def stream_new_blocks_ahead(
        self,
        blockchain: str,
        start_block_hash_or_height: Union[int, str] = "latest",
        sleep_interval_seconds: float = 10,
        n_blocks_ahead: int = 10,
    ) -> Iterable[Block]:
        if not self._executor:
            raise Exception("Works only with executor")

        if n_blocks_ahead < 1:
            raise ValueError(f"`n_blocks_ahead` must be greater than 0. Got `{n_blocks_ahead}`")

        block = self.get_block(blockchain=blockchain, hash_or_height=start_block_hash_or_height)
        next_block_height = block.height + 1
        last_block_height = next_block_height + n_blocks_ahead + 1
        yield block

        block_tasks = [
            self._executor.submit(self.get_block, blockchain, height)
            for height in range(next_block_height, last_block_height)
        ]

        while True:
            block_task = block_tasks.pop(0)
            try:
                yield block_task.result()
            except HTTPError as exception:
                if exception.response.status_code == 404:
                    self._handle_404_during_block_streaming(blockchain, next_block_height, sleep_interval_seconds)
                    block_tasks.insert(0, self._executor.submit(self.get_block, blockchain, next_block_height))
                    continue

                raise

            block_tasks.append(self._executor.submit(self.get_block, blockchain, last_block_height))
            next_block_height += 1
            last_block_height += 1

    def stream_new_transactions(
        self,
        blockchain: str,
        start_block_hash_or_height: Union[int, str] = "latest",
        sleep_interval_seconds: float = 10,
    ) -> Iterable[Tuple[Block, Transaction]]:
        new_blocks = self.stream_new_blocks(
            blockchain=blockchain,
            start_block_hash_or_height=start_block_hash_or_height,
            sleep_interval_seconds=sleep_interval_seconds,
        )
        for block in new_blocks:
            block_transactions = self.get_all_transactions_of_block(blockchain=blockchain, block=block)
            for transaction in block_transactions:
                yield block, transaction

    def get_tokens(self, blockchain: str) -> Dict[str, Token]:
        if self._tokens.get(blockchain) is None:
            self._refresh_tokens(blockchain)

        return self._tokens[blockchain]

    def scale_token_value(self, blockchain: str, contract_address: str, value: Decimal) -> Decimal:
        token = self.get_tokens(blockchain)[contract_address]
        return value * token.scaling

    def _get(self, blockchain: str, endpoint: str) -> dict:  # type: ignore[return]
        for iteration in count(start=1):
            try:
                response = self._session.get(
                    f"https://{blockchain}.stocra.com/v1.0/{endpoint}",
                    allow_redirects=False,
                    headers=self.headers,
                )
                response.raise_for_status()
                return cast(dict, response.json())
            except RequestException as exception:
                error = StocraHTTPError(endpoint=endpoint, iteration=iteration, exception=exception)
                if self._should_continue(error):
                    continue

                raise

    def _should_continue(self, error: StocraHTTPError) -> bool:
        if not self._error_handlers:
            return False

        for error_handler in self._error_handlers:
            retry = error_handler(error)
            if retry:
                return True

        return False

    @classmethod
    def _handle_404_during_block_streaming(
        cls, blockchain: str, block_height: int, sleep_interval_seconds: float
    ) -> None:
        logger.debug(
            "%s: stream_new_blocks %s: 404, sleeping for %d seconds",
            blockchain,
            block_height,
            sleep_interval_seconds,
        )
        sleep(sleep_interval_seconds)

    def _refresh_tokens(self, blockchain: str) -> None:
        tokens = self._get(blockchain, "tokens")
        self._tokens[blockchain] = {contract_address: Token(**token) for contract_address, token in tokens.items()}



import abc
from typing import Dict, List, Optional

from stocra.models import ErrorHandler, Token


class StocraBase(abc.ABC):
    _api_key: Optional[str] = None
    _error_handlers: Optional[List[ErrorHandler]] = None
    _tokens: Dict[str, Dict[str, Token]] = dict()

    def __init__(
        self,
        api_key: Optional[str] = None,
        error_handlers: Optional[List[ErrorHandler]] = None,
    ) -> None:
        self._api_key = api_key
        self._error_handlers = error_handlers

    @property
    def headers(self) -> dict:
        if self._api_key:
            return dict(Authorization=f"Bearer {self._api_key}")

        return dict()



from dataclasses import dataclass
from decimal import Decimal
from enum import Enum, unique
from typing import Any, Awaitable, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, root_validator, validator

Address = str
TransactionHash = str
OutputIndex = int


class Amount(BaseModel):
    value: Decimal
    currency_symbol: str

    def __add__(self, other: Any) -> "Amount":
        if isinstance(other, Amount):
            if self.currency_symbol != other.currency_symbol:
                raise ValueError(
                    f"Amounts have different currencies: {self.currency_symbol} != {other.currency_symbol}"
                )

            return Amount(value=self.value + other.value, currency_symbol=self.currency_symbol)

        raise TypeError(f"Cannot add Amount and {type(other)}")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Amount):
            raise TypeError(f"Cannot compare Amount and {type(other)}")

        if self.value != other.value:
            return False

        if self.currency_symbol != other.currency_symbol:
            return False

        return True


class TransactionPointer(BaseModel):
    transaction_hash: TransactionHash
    output_index: OutputIndex


class Input(BaseModel):
    address: Optional[Address] = None
    amount: Optional[Amount] = None
    transaction_pointer: Optional[TransactionPointer] = None

    @classmethod
    @root_validator
    def validate_address_or_pointer(cls, values: Dict) -> Dict:  # pylint: disable=no-self-argument
        if not (values["address"] or values["transaction_pointer"]):
            raise ValueError("Either address or transaction pointer must be specified")

        return values


class Output(BaseModel):
    address: Address
    amount: Amount


class Transaction(BaseModel):
    hash: TransactionHash
    inputs: List[Input]
    outputs: List[Output]
    fee: Amount

    @classmethod
    @validator("fee")
    def validate_fee(cls, value: Amount) -> Amount:
        if value.value < 0:
            raise ValueError(f"Fee cannot be negative! Got {value.value}")

        return value


class Block(BaseModel):
    height: int
    hash: str
    timestamp_ms: int
    transactions: List[str] = []

    @classmethod
    @validator("timestamp_ms")
    def validate_timestamp_ms(cls, value: int) -> int:  # pylint: disable=no-self-argument
        # I don't really care if this runs even after 2286
        if len(str(value)) != 13:
            raise ValueError("Timestamp must be in miliseconds")

        return value


class Currency(BaseModel):
    symbol: str
    name: str

    class Config:
        frozen = True


@unique
class TokenType(Enum):
    ERC20 = "ERC20"


class Token(BaseModel):
    currency: Currency
    scaling: Decimal
    type: TokenType


@dataclass(frozen=True)
class StocraHTTPError:
    endpoint: str
    iteration: int
    exception: Exception


ErrorHandler = Callable[[StocraHTTPError], Union[bool, Awaitable[bool]]]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stocra', 'stocra.asynchronous', 'stocra.synchronous']

package_data = \
{'': ['*']}

install_requires = \
['pydantic']

extras_require = \
{'asynchronous': ['aiohttp'], 'synchronous': ['requests']}

setup_kwargs = {
    'name': 'stocra',
    'version': '0.1.7',
    'description': 'Stocra.com python sdk',
    'long_description': '# Stocra Python SDK\n- [SDK API documentation](https://stocra.github.io/sdk-python/)\n  - [Models](https://stocra.github.io/sdk-python/stocra/models.html)\n  - [Synchronous client](https://stocra.github.io/sdk-python/stocra/synchronous/client.html)\n  - [Asynchronous client](https://stocra.github.io/sdk-python/stocra/asynchronous/client.html)\n- [Using synchronous client](#synchronous-client)\n- [Using asynchronous client](#asynchronous-client)\n- [Error handlers](#error-handlers)\n\n## Synchronous client\n### Install\n```bash\npip install stocra[synchronous]\n# or\npoetry add stocra --extras synchronous\n```\n### Usage\n```python\nfrom concurrent.futures import ThreadPoolExecutor\nfrom decimal import Decimal\n\nfrom requests import Session\nfrom requests.adapters import HTTPAdapter\nfrom stocra.synchronous.client import Stocra\nfrom stocra.synchronous.error_handlers import retry_on_too_many_requests, retry_on_service_unavailable\n\nadapter = HTTPAdapter(pool_connections=100, pool_maxsize=100)\nsession = Session()\nsession.mount(\'https://\', adapter)\nstocra_client = Stocra(\n    api_key="<api-key>", # optional\n    session=session, # optional\n    executor=ThreadPoolExecutor(), # optional\n    error_handlers=[ \n        retry_on_service_unavailable,\n        retry_on_too_many_requests,\n    ] # optional\n)\n\n# stream new blocks\nfor block in stocra_client.stream_new_blocks(blockchain="ethereum"):\n    print(block)\n\n# stream new blocks, load new blocks in the background for faster processing. \n# Works only if executor was provided during instantiation.\nfor block in stocra_client.stream_new_blocks_ahead(blockchain="ethereum"):\n    print(block)\n    \n# stream new transactions\nfor block, transaction in stocra_client.stream_new_transactions(blockchain="ethereum"):\n    print(block.height, transaction.hash)\n    \n# get one block\nblock = stocra_client.get_block(blockchain="bitcoin", hash_or_height=57043)\n\n# get one transaction\ntransaction = stocra_client.get_transaction(\n    blockchain="bitcoin", \n    transaction_hash="a1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d"\n)\n\n# get all transactions in block\ntransactions = stocra_client.get_all_transactions_of_block(blockchain="bitcoin", block=block) \nfor transaction in transactions:\n    print(transaction)\n    \n# scale token value\nvalue = stocra_client.scale_token_value(\n    "ethereum", \n    "0xa49ded8b4607f958003e0d87d7f2d2f69bcadd41",  # USDT\n    Decimal("34500000000000000000000000") # raw value in token transfer\n)\n```\n## Asynchronous client\n### Install\n```bash\npip install stocra[asynchronous]\n# or\npoetry add stocra --extras asynchronous\n```\n### Usage\n```python\nfrom asyncio import Semaphore\nfrom decimal import Decimal\n\nfrom aiohttp import ClientSession\nfrom stocra.asynchronous.client import Stocra\nfrom stocra.asynchronous.error_handlers import retry_on_too_many_requests, retry_on_service_unavailable\n\nsession = ClientSession()\nstocra_client = Stocra(\n    api_key="<api-key>", # optional\n    session=session, # optional\n    semaphore=Semaphore(50), # optional\n    error_handlers=[\n        retry_on_service_unavailable,\n        retry_on_too_many_requests,\n    ] # optional\n)\n# stream new transactions\nasync for block, transaction in stocra_client.stream_new_transactions(blockchain="ethereum"):\n    print(block.height, transaction.hash)\n\n# stream new blocks and always load next 5 blocks in the background.\n# useful when you need to parse multiple blocks in short time span\nasync for block in stocra_client.stream_new_blocks(blockchain="ethereum", n_blocks_ahead=5):\n    print(block)\n\n# get one block\nblock = await stocra_client.get_block(\n    blockchain="bitcoin",\n    hash_or_height="00000000152340ca42227603908689183edc47355204e7aca59383b0aaac1fd8"\n)\n\n# get one transaction\ntransaction = await stocra_client.get_transaction(\n    blockchain="bitcoin",\n    transaction_hash="a1075db55d416d3ca199f55b6084e2115b9345e16c5cf302fc80e9d5fbf5d48d", \n)\n\n# get all transactions in block\ntransactions = stocra_client.get_all_transactions_of_block(blockchain="bitcoin", block=block)\nasync for transaction in transactions:\n    print(transaction)\n\n# scale token value\nvalue = await stocra_client.scale_token_value(\n    "ethereum", \n    "0xa49ded8b4607f958003e0d87d7f2d2f69bcadd41",  # USDT\n    Decimal("34500000000000000000000000") # raw value in token transfer\n)\n\n```\n## Error handlers\nError handlers are functions that are called after a request fails. \nThey receive single argument, [StocraHTTPError](https://stocra.github.io/sdk-python/stocra/models.html#StocraHTTPError) \nand return boolean indicating whether to retry request (`True`) or raise (`False`).\n\nError handler signature: `ErrorHandler = Callable[[StocraHTTPError], Union[bool, Awaitable[bool]]]`\n\nNo errors handlers are used by default although there are two already defined for both sync and async version: \n- synchronous error handlers: [stocra.synchronous.error_handlers](https://stocra.github.io/sdk-python/stocra/synchronous/error_handlers.html)\n- of asynchronous error handlers: [stocra.asynchronous.error_handlers](https://stocra.github.io/sdk-python/stocra/asynchronous/error_handlers.html)\n',
    'author': 'Lukáš Vokráčko',
    'author_email': 'lukas@vokracko.cz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://stocra.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import asyncio
import logging
from asyncio import Semaphore
from contextlib import asynccontextmanager
from decimal import Decimal
from itertools import count
from typing import (
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from aiohttp import ClientError, ClientResponseError, ClientSession

from stocra.base_client import StocraBase
from stocra.models import Block, ErrorHandler, StocraHTTPError, Token, Transaction

logger = logging.getLogger("stocra")


class Stocra(StocraBase):
    _session: ClientSession
    _semaphore: Optional[Semaphore]
    _error_handlers: List[ErrorHandler]

    def __init__(
        self,
        api_key: Optional[str] = None,
        session: Optional[ClientSession] = None,
        semaphore: Optional[Semaphore] = None,
        error_handlers: Optional[List[ErrorHandler]] = None,
    ):
        super().__init__(
            api_key=api_key,
            error_handlers=error_handlers,
        )

        self._session = session or ClientSession()
        self._semaphore = semaphore

    async def close(self) -> None:
        await self._session.close()

    async def get_block(self, blockchain: str, hash_or_height: Union[str, int] = "latest") -> Block:
        logger.debug("%s: get_block %s", blockchain, hash_or_height)
        async with self._with_semaphore():
            block_json = await self._get(blockchain=blockchain, endpoint=f"blocks/{hash_or_height}")
            return Block(**block_json)

    async def get_transaction(self, blockchain: str, transaction_hash: str) -> Transaction:
        logger.debug("%s: get_transaction %s", blockchain, transaction_hash)
        async with self._with_semaphore():
            transaction_json = await self._get(blockchain=blockchain, endpoint=f"transactions/{transaction_hash}")
            return Transaction(**transaction_json)

    async def get_all_transactions_of_block(self, blockchain: str, block: Block) -> AsyncIterable[Transaction]:
        logger.debug("%s: get_all_transactions %s", blockchain, block.height)
        transaction_tasks = []

        for transaction_hash in block.transactions:
            task = asyncio.create_task(self.get_transaction(blockchain=blockchain, transaction_hash=transaction_hash))
            transaction_tasks.append(task)

        for completed_task in asyncio.as_completed(transaction_tasks):
            yield await completed_task

    async def stream_new_blocks(
        self,
        blockchain: str,
        start_block_hash_or_height: Union[int, str] = "latest",
        sleep_interval_seconds: float = 10,
        n_blocks_ahead: int = 1,
    ) -> AsyncIterable[Block]:
        if n_blocks_ahead < 1:
            raise ValueError(f"`n_blocks_ahead` must be greater than 0. Got `{n_blocks_ahead}`")

        block = await self.get_block(blockchain=blockchain, hash_or_height=start_block_hash_or_height)
        first_block_to_load_height = block.height + 1
        last_block_to_load_height = first_block_to_load_height + n_blocks_ahead + 1
        yield block

        block_tasks = [
            asyncio.create_task(self.get_block(blockchain, height))
            for height in range(first_block_to_load_height, last_block_to_load_height)
        ]

        while True:
            block_task = block_tasks.pop(0)
            try:
                await asyncio.wait_for(block_task, timeout=None)
                yield block_task.result()
            except ClientResponseError as exception:
                if exception.status == 404:
                    logger.debug(
                        "%s: stream_new_blocks_ahead %s: 404, sleeping for %d seconds",
                        blockchain,
                        first_block_to_load_height,
                        sleep_interval_seconds,
                    )
                    await asyncio.sleep(sleep_interval_seconds)
                    block_tasks.insert(0, asyncio.create_task(self.get_block(blockchain, first_block_to_load_height)))
                    continue

                raise

            block_tasks.append(asyncio.create_task(self.get_block(blockchain, last_block_to_load_height)))
            first_block_to_load_height += 1
            last_block_to_load_height += 1

    async def stream_new_transactions(
        self,
        blockchain: str,
        start_block_hash_or_height: Union[int, str] = "latest",
        sleep_interval_seconds: float = 10,
    ) -> AsyncIterable[Tuple[Block, Transaction]]:
        new_blocks = self.stream_new_blocks(
            blockchain=blockchain,
            start_block_hash_or_height=start_block_hash_or_height,
            sleep_interval_seconds=sleep_interval_seconds,
        )
        async for block in new_blocks:
            block_transactions = self.get_all_transactions_of_block(blockchain=blockchain, block=block)
            async for transaction in block_transactions:
                yield block, transaction

    async def get_tokens(self, blockchain: str) -> Dict[str, Token]:
        if self._tokens.get(blockchain) is None:
            await self._refresh_tokens(blockchain)

        return self._tokens[blockchain]

    async def scale_token_value(self, blockchain: str, contract_address: str, value: Decimal) -> Decimal:
        tokens = await self.get_tokens(blockchain)
        token = tokens[contract_address]
        return value * token.scaling

    async def _acquire(self) -> None:
        if self._semaphore:
            await self._semaphore.acquire()

    def _release(self) -> None:
        if self._semaphore:
            self._semaphore.release()

    @asynccontextmanager
    async def _with_semaphore(self) -> AsyncGenerator[None, None]:
        await self._acquire()
        try:
            yield
        finally:
            self._release()

    async def _get(self, blockchain: str, endpoint: str) -> dict:  # type: ignore[return]
        for iteration in count(start=1):
            try:
                response = await self._session.get(
                    f"https://{blockchain}.stocra.com/v1.0/{endpoint}",
                    raise_for_status=True,
                    allow_redirects=False,
                    headers=self.headers,
                )
                return cast(dict, await response.json())
            except ClientError as exception:
                error = StocraHTTPError(endpoint=endpoint, iteration=iteration, exception=exception)
                if await self._should_continue(error):
                    continue

                raise

    async def _should_continue(self, error: StocraHTTPError) -> bool:
        if not self._error_handlers:
            return False

        for error_handler in self._error_handlers:
            retry = await cast(Awaitable[bool], error_handler(error))
            if retry:
                return True

        return False

    async def _refresh_tokens(self, blockchain: str) -> None:
        tokens = await self._get(blockchain, "tokens")
        self._tokens[blockchain] = {contract_address: Token(**token) for contract_address, token in tokens.items()}
